@@ -1,5 +1,6 @@
 #include <assert.h>
 #include <errno.h>
+#include <math.h>
 #include <netdb.h>
 #include <netinet/in.h>
 #include <poll.h>
@@ -20,7 +21,6 @@
 #include "httpinput.h"
 #include "log.h"
 #include "metacube2.h"
-#include "mutexlock.h"
 #include "parse.h"
 #include "serverpool.h"
 #include "state.pb.h"
@@ -49,22 +49,14 @@ extern ServerPool *servers;
 HTTPInput::HTTPInput(const string &url, Input::Encoding encoding)
 	: state(NOT_CONNECTED),
 	  url(url),
-	  encoding(encoding),
-	  has_metacube_header(false),
-	  sock(-1),
-	  num_connection_attempts(0),
-	  suppress_logging(false)
+	  encoding(encoding)
 {
-	pthread_mutex_init(&stats_mutex, NULL);
 	stats.url = url;
 	stats.bytes_received = 0;
 	stats.data_bytes_received = 0;
 	stats.metadata_bytes_received = 0;
 	stats.connect_time = -1;
 	stats.latency_sec = HUGE_VAL;
-
-	last_verbose_connection.tv_sec = -3600;
-	last_verbose_connection.tv_nsec = 0;
 }
 
 HTTPInput::HTTPInput(const InputProto &serialized)
@@ -79,9 +71,7 @@ HTTPInput::HTTPInput(const InputProto &serialized)
 	  http_header(serialized.http_header()),
 	  stream_header(serialized.stream_header()),
 	  has_metacube_header(serialized.has_metacube_header()),
-	  sock(serialized.sock()),
-	  num_connection_attempts(0),
-	  suppress_logging(false)
+	  sock(serialized.sock())
 {
 	pending_data.resize(serialized.pending_data().size());
 	memcpy(&pending_data[0], serialized.pending_data().data(), serialized.pending_data().size());
@@ -89,7 +79,6 @@ HTTPInput::HTTPInput(const InputProto &serialized)
 	string protocol, user;
 	parse_url(url, &protocol, &user, &host, &port, &path);  // Don't care if it fails.
 
-	pthread_mutex_init(&stats_mutex, NULL);
 	stats.url = url;
 	stats.bytes_received = serialized.bytes_received();
 	stats.data_bytes_received = serialized.data_bytes_received();
@@ -97,7 +86,7 @@ HTTPInput::HTTPInput(const InputProto &serialized)
 	if (serialized.has_connect_time()) {
 		stats.connect_time = serialized.connect_time();
 	} else {
-		stats.connect_time = time(NULL);
+		stats.connect_time = time(nullptr);
 	}
 	if (serialized.has_latency_sec()) {
 		stats.latency_sec = serialized.latency_sec();
@@ -116,7 +105,7 @@ void HTTPInput::close_socket()
 		sock = -1;
 	}
 
-	MutexLock lock(&stats_mutex);
+	lock_guard<mutex> lock(stats_mutex);
 	stats.connect_time = -1;
 }
 
@@ -151,7 +140,7 @@ InputProto HTTPInput::serialize() const
 int HTTPInput::lookup_and_connect(const string &host, const string &port)
 {
 	addrinfo *ai;
-	int err = getaddrinfo(host.c_str(), port.c_str(), NULL, &ai);
+	int err = getaddrinfo(host.c_str(), port.c_str(), nullptr, &ai);
 	if (err != 0) {
 		if (!suppress_logging) {
 			log(WARNING, "[%s] Lookup of '%s' failed (%s).",
@@ -194,7 +183,7 @@ int HTTPInput::lookup_and_connect(const string &host, const string &port)
 
 		// Wait for the connect to complete, or an error to happen.
 		for ( ;; ) {
-			bool complete = wait_for_activity(sock, POLLIN | POLLOUT, NULL);
+			bool complete = wait_for_activity(sock, POLLIN | POLLOUT, nullptr);
 			if (should_stop()) {
 				safe_close(sock);
 				return -1;
@@ -260,47 +249,38 @@ bool HTTPInput::parse_response(const string &request)
 		return false;
 	}
 
-	multimap<string, string> parameters = extract_headers(lines, url);
+	HTTPHeaderMultimap parameters = extract_headers(lines, url);
 
 	// Remove “Content-encoding: metacube”.
-	// TODO: Make case-insensitive.
-	multimap<string, string>::iterator encoding_it =
-		parameters.find("Content-encoding");
+	const auto encoding_it = parameters.find("Content-Encoding");
 	if (encoding_it != parameters.end() && encoding_it->second == "metacube") {
 		parameters.erase(encoding_it);
 	}
 
 	// Change “Server: foo” to “Server: metacube/0.1 (reflecting: foo)”
-	// TODO: Make case-insensitive.
 	// XXX: Use a Via: instead?
 	if (parameters.count("Server") == 0) {
 		parameters.insert(make_pair("Server", SERVER_IDENTIFICATION));
 	} else {
-		for (multimap<string, string>::iterator it = parameters.begin();
-		     it != parameters.end();
-		     ++it) {
-			if (it->first != "Server") {
+		for (auto &key_and_value : parameters) {
+			if (key_and_value.first != "Server") {
 				continue;
 			}
-			it->second = SERVER_IDENTIFICATION " (reflecting: " + it->second + ")";
+			key_and_value.second = SERVER_IDENTIFICATION " (reflecting: " + key_and_value.second + ")";
 		}
 	}
 
-	// Set “Connection: close”.
-	// TODO: Make case-insensitive.
+	// Erase “Connection: close”; we'll set it on the sending side if needed.
 	parameters.erase("Connection");
-	parameters.insert(make_pair("Connection", "close"));
 
 	// Construct the new HTTP header.
 	http_header = "HTTP/1.0 200 OK\r\n";
-	for (multimap<string, string>::iterator it = parameters.begin();
-	     it != parameters.end();
-	     ++it) {
-		http_header.append(it->first + ": " + it->second + "\r\n");
+	for (const auto &key_and_value : parameters) {
+		http_header.append(key_and_value.first + ": " + key_and_value.second + "\r\n");
 	}
 
-	for (size_t i = 0; i < stream_indices.size(); ++i) {
-		servers->set_header(stream_indices[i], http_header, stream_header);
+	for (int stream_index : stream_indices) {
+		servers->set_header(stream_index, http_header, stream_header);
 	}
 
 	return true;
@@ -358,8 +338,8 @@ void HTTPInput::do_work()
 			response.clear();
 			pending_data.clear();
 			has_metacube_header = false;
-			for (size_t i = 0; i < stream_indices.size(); ++i) {
-				servers->set_header(stream_indices[i], "", "");
+			for (int stream_index : stream_indices) {
+				servers->set_header(stream_index, "", "");
 			}
 
 			{
@@ -410,8 +390,8 @@ void HTTPInput::do_work()
 					request_bytes_sent = 0;
 				}
 
-				MutexLock lock(&stats_mutex);
-				stats.connect_time = time(NULL);
+				lock_guard<mutex> lock(stats_mutex);
+				stats.connect_time = time(nullptr);
 				clock_gettime(CLOCK_MONOTONIC_COARSE, &last_activity);
 			}
 			break;
@@ -479,7 +459,7 @@ void HTTPInput::do_work()
 			if (status == RP_EXTRA_DATA) {
 				char *ptr = static_cast<char *>(
 					memmem(response.data(), response.size(), "\r\n\r\n", 4));
-				assert(ptr != NULL);
+				assert(ptr != nullptr);
 				extra_data = string(ptr + 4, &response[0] + response.size());
 				response.resize(ptr - response.data());
 			}
@@ -580,13 +560,13 @@ void HTTPInput::do_work()
 void HTTPInput::process_data(char *ptr, size_t bytes)
 {
 	{
-		MutexLock mutex(&stats_mutex);
+		lock_guard<mutex> lock(stats_mutex);
 		stats.bytes_received += bytes;
 	}
 
 	if (encoding == Input::INPUT_ENCODING_RAW) {
-		for (size_t i = 0; i < stream_indices.size(); ++i) {
-			servers->add_data(stream_indices[i], ptr, bytes, /*metacube_flags=*/0);
+		for (int stream_index : stream_indices) {
+			servers->add_data(stream_index, ptr, bytes, /*metacube_flags=*/0, /*pts=*/RationalPTS());
 		}
 		return;
 	}
@@ -606,7 +586,7 @@ void HTTPInput::process_data(char *ptr, size_t bytes)
 			char *ptr = static_cast<char *>(
 				memmem(pending_data.data(), pending_data.size(),
 				       METACUBE2_SYNC, strlen(METACUBE2_SYNC)));
-			if (ptr == NULL) {
+			if (ptr == nullptr) {
 				// OK, so we didn't find the sync marker. We know then that
 				// we do not have the _full_ marker in the buffer, but we
 				// could have N-1 bytes. Drop everything before that,
@@ -659,26 +639,27 @@ void HTTPInput::process_data(char *ptr, size_t bytes)
 		// TODO: Keep metadata when sending on to other Metacube users.
 		if (flags & METACUBE_FLAGS_METADATA) {
 			{
-				MutexLock lock(&stats_mutex);
+				lock_guard<mutex> lock(stats_mutex);
 				stats.metadata_bytes_received += size;
 			}
 			process_metacube_metadata_block(hdr, pending_data.data() + sizeof(hdr), size);
 		} else {
 			// Send this block on to the servers.
 			{
-				MutexLock lock(&stats_mutex);
+				lock_guard<mutex> lock(stats_mutex);
 				stats.data_bytes_received += size;
 			}
 			char *inner_data = pending_data.data() + sizeof(metacube2_block_header);
 			if (flags & METACUBE_FLAGS_HEADER) {
 				stream_header = string(inner_data, inner_data + size);
-				for (size_t i = 0; i < stream_indices.size(); ++i) {
-					servers->set_header(stream_indices[i], http_header, stream_header);
+				for (int stream_index : stream_indices) {
+					servers->set_header(stream_index, http_header, stream_header);
 				}
 			}
-			for (size_t i = 0; i < stream_indices.size(); ++i) {
-				servers->add_data(stream_indices[i], inner_data, size, flags);
+			for (int stream_index : stream_indices) {
+				servers->add_data(stream_index, inner_data, size, flags, next_block_pts);
 			}
+			next_block_pts = RationalPTS();
 		}
 
 		// Consume the block. This isn't the most efficient way of dealing with things
@@ -708,7 +689,7 @@ void HTTPInput::add_destination(int stream_index)
 
 InputStats HTTPInput::get_stats() const
 {
-	MutexLock lock(&stats_mutex);
+	lock_guard<mutex> lock(stats_mutex);
 	return stats;
 }
 
@@ -721,25 +702,36 @@ void HTTPInput::process_metacube_metadata_block(const metacube2_block_header &hd
 	}
 
 	uint64_t type = be64toh(*(const uint64_t *)payload);
-	if (type != METACUBE_METADATA_TYPE_ENCODER_TIMESTAMP) {
-		// Unknown metadata block, ignore.
+	if (type == METACUBE_METADATA_TYPE_ENCODER_TIMESTAMP) {
+		timespec now;
+		clock_gettime(CLOCK_REALTIME, &now);
+
+		const metacube2_timestamp_packet *pkt = (const metacube2_timestamp_packet *)payload;
+		if (payload_size != sizeof(*pkt)) {
+			log(WARNING, "[%s] Metacube timestamp block of wrong size (%d bytes); ignoring.",
+				url.c_str(), payload_size);
+			return;
+		}
+
+		double elapsed = now.tv_sec - be64toh(pkt->tv_sec) +
+			1e-9 * (now.tv_nsec - long(be64toh(pkt->tv_nsec)));
+		{
+			lock_guard<mutex> lock(stats_mutex);
+			stats.latency_sec = elapsed;
+		}
+	} else if (type == METACUBE_METADATA_TYPE_NEXT_BLOCK_PTS) {
+		const metacube2_pts_packet *pkt = (const metacube2_pts_packet *)payload;
+		if (payload_size != sizeof(*pkt)) {
+			log(WARNING, "[%s] Metacube pts block of wrong size (%d bytes); ignoring.",
+				url.c_str(), payload_size);
+			return;
+		}
+		next_block_pts.pts = be64toh(pkt->pts);
+		next_block_pts.timebase_num = be64toh(pkt->timebase_num);
+		next_block_pts.timebase_den = be64toh(pkt->timebase_den);
+	} else {
+		// Unknown metadata block, ignore
+		log(INFO, "[%s] Metadata block %llu\n", url.c_str(), type);
 		return;
 	}
-
-	timespec now;
-	clock_gettime(CLOCK_REALTIME, &now);
-
-	const metacube2_timestamp_packet *pkt = (const metacube2_timestamp_packet *)payload;
-	if (payload_size != sizeof(*pkt)) {
-		log(WARNING, "[%s] Metacube timestamp block of wrong size (%d bytes); ignoring.",
-			url.c_str(), payload_size);
-		return;
-	}
-
-	double elapsed = now.tv_sec - be64toh(pkt->tv_sec) +
-		1e-9 * (now.tv_nsec - long(be64toh(pkt->tv_nsec)));
-	{
-		MutexLock lock(&stats_mutex);
-		stats.latency_sec = elapsed;
-	}
 }
