@@ -1,6 +1,8 @@
 #include <assert.h>
 #include <errno.h>
+#include <inttypes.h>
 #include <limits.h>
+#include <math.h>
 #include <netinet/in.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -13,28 +15,33 @@
 
 #include "log.h"
 #include "metacube2.h"
-#include "mutexlock.h"
 #include "state.pb.h"
 #include "stream.h"
 #include "util.h"
 
 using namespace std;
 
-Stream::Stream(const string &url, size_t backlog_size, size_t prebuffering_bytes, Encoding encoding, Encoding src_encoding)
+Stream::Stream(const string &url,
+               size_t backlog_size,
+               uint64_t prebuffering_bytes,
+               Encoding encoding,
+               Encoding src_encoding,
+               unsigned hls_frag_duration,
+               size_t hls_backlog_margin,
+               const std::string &allow_origin)
 	: url(url),
 	  encoding(encoding),
 	  src_encoding(src_encoding),
+	  allow_origin(allow_origin),
 	  data_fd(make_tempfile("")),
-          backlog_size(backlog_size),
+	  backlog_size(backlog_size),
 	  prebuffering_bytes(prebuffering_bytes),
-	  bytes_received(0),
-	  pacing_rate(~0U)
+	  hls_frag_duration(hls_frag_duration),
+	  hls_backlog_margin(hls_backlog_margin)
 {
 	if (data_fd == -1) {
 		exit(1);
 	}
-
-	pthread_mutex_init(&queued_data_mutex, NULL);
 }
 
 Stream::~Stream()
@@ -51,16 +58,15 @@ Stream::Stream(const StreamProto &serialized, int data_fd)
 	  encoding(Stream::STREAM_ENCODING_RAW),  // Will be changed later.
 	  data_fd(data_fd),
 	  backlog_size(serialized.backlog_size()),
-	  prebuffering_bytes(serialized.prebuffering_bytes()),
 	  bytes_received(serialized.bytes_received()),
-	  pacing_rate(~0U)
+	  first_fragment_index(serialized.first_fragment_index()),
+	  discontinuity_counter(serialized.discontinuity_counter())
 {
 	if (data_fd == -1) {
 		exit(1);
 	}
 
-	for (int i = 0; i < serialized.suitable_starting_point_size(); ++i) {
-		ssize_t point = serialized.suitable_starting_point(i);
+	for (ssize_t point : serialized.suitable_starting_point()) {
 		if (point == -1) {
 			// Can happen when upgrading from before 1.1.3,
 			// where this was an optional field with -1 signifying
@@ -70,7 +76,9 @@ Stream::Stream(const StreamProto &serialized, int data_fd)
 		suitable_starting_points.push_back(point);
 	}
 
-	pthread_mutex_init(&queued_data_mutex, NULL);
+	for (const FragmentStartProto &fragment : serialized.fragment()) {
+		fragments.push_back(FragmentStart { size_t(fragment.byte_position()), fragment.pts() });
+	}
 }
 
 StreamProto Stream::serialize()
@@ -80,11 +88,18 @@ StreamProto Stream::serialize()
 	serialized.set_stream_header(stream_header);
 	serialized.add_data_fds(data_fd);
 	serialized.set_backlog_size(backlog_size);
-	serialized.set_prebuffering_bytes(prebuffering_bytes);
 	serialized.set_bytes_received(bytes_received);
-	for (size_t i = 0; i < suitable_starting_points.size(); ++i) {
-		serialized.add_suitable_starting_point(suitable_starting_points[i]);
+	for (size_t point : suitable_starting_points) {
+		serialized.add_suitable_starting_point(point);
 	}
+	for (const FragmentStart &fragment : fragments) {
+		FragmentStartProto *proto = serialized.add_fragment();
+		proto->set_byte_position(fragment.byte_position);
+		proto->set_pts(fragment.pts);
+	}
+	serialized.set_first_fragment_index(first_fragment_index);
+	serialized.set_discontinuity_counter(discontinuity_counter);
+
 	serialized.set_url(url);
 	data_fd = -1;
 	return serialized;
@@ -178,6 +193,7 @@ vector<Stream::DataElement> remove_iovecs(const vector<Stream::DataElement> &dat
 			data_element.data.iov_base = reinterpret_cast<char *>(data[i].data.iov_base) + bytes_wanted;
 			data_element.data.iov_len = data[i].data.iov_len - bytes_wanted;
 			data_element.metacube_flags = METACUBE_FLAGS_NOT_SUITABLE_FOR_STREAM_START;
+			data_element.pts = RationalPTS();
 			ret.push_back(data_element);
 			bytes_wanted = 0;
 		}
@@ -222,9 +238,16 @@ void Stream::remove_obsolete_starting_points()
 	       bytes_received - suitable_starting_points[0] > backlog_size) {
 		suitable_starting_points.pop_front();
 	}
+	assert(backlog_size >= hls_backlog_margin);
+	while (!fragments.empty() &&
+	       bytes_received - fragments[0].byte_position > (backlog_size - hls_backlog_margin)) {
+		fragments.pop_front();
+		++first_fragment_index;
+		clear_hls_playlist_cache();
+	}
 }
 
-void Stream::add_data_deferred(const char *data, size_t bytes, uint16_t metacube_flags)
+void Stream::add_data_deferred(const char *data, size_t bytes, uint16_t metacube_flags, const RationalPTS &pts)
 {
 	// For regular output, we don't want to send the client twice
 	// (it's already sent out together with the HTTP header).
@@ -236,12 +259,27 @@ void Stream::add_data_deferred(const char *data, size_t bytes, uint16_t metacube
 		return;
 	}
 
-	MutexLock lock(&queued_data_mutex);
+	lock_guard<mutex> lock(queued_data_mutex);
 
 	DataElement data_element;
 	data_element.metacube_flags = metacube_flags;
+	data_element.pts = pts;
 
 	if (encoding == Stream::STREAM_ENCODING_METACUBE) {
+		// Construct a PTS metadata block. (We'll avoid sending it out
+		// if we don't have a valid PTS.)
+		metacube2_pts_packet pts_packet;
+		pts_packet.type = htobe64(METACUBE_METADATA_TYPE_NEXT_BLOCK_PTS);
+		pts_packet.pts = htobe64(pts.pts);
+		pts_packet.timebase_num = htobe64(pts.timebase_num);
+		pts_packet.timebase_den = htobe64(pts.timebase_den);
+
+		metacube2_block_header pts_hdr;
+		memcpy(pts_hdr.sync, METACUBE2_SYNC, sizeof(pts_hdr.sync));
+		pts_hdr.size = htonl(sizeof(pts_packet));
+		pts_hdr.flags = htons(METACUBE_FLAGS_METADATA);
+		pts_hdr.csum = htons(metacube2_compute_crc(&pts_hdr));
+
 		// Add a Metacube block header before the data.
 		metacube2_block_header hdr;
 		memcpy(hdr.sync, METACUBE2_SYNC, sizeof(hdr.sync));
@@ -249,11 +287,23 @@ void Stream::add_data_deferred(const char *data, size_t bytes, uint16_t metacube
 		hdr.flags = htons(metacube_flags);
 		hdr.csum = htons(metacube2_compute_crc(&hdr));
 
-		data_element.data.iov_base = new char[bytes + sizeof(hdr)];
 		data_element.data.iov_len = bytes + sizeof(hdr);
+		if (pts.timebase_num != 0) {
+			data_element.data.iov_len += sizeof(pts_hdr) + sizeof(pts_packet);
+		}
+		data_element.data.iov_base = new char[data_element.data.iov_len];
 
-		memcpy(data_element.data.iov_base, &hdr, sizeof(hdr));
-		memcpy(reinterpret_cast<char *>(data_element.data.iov_base) + sizeof(hdr), data, bytes);
+		char *ptr = reinterpret_cast<char *>(data_element.data.iov_base);
+		if (pts.timebase_num != 0) {
+			memcpy(ptr, &pts_hdr, sizeof(pts_hdr));
+			ptr += sizeof(pts_hdr);
+			memcpy(ptr, &pts_packet, sizeof(pts_packet));
+			ptr += sizeof(pts_packet);
+		}
+
+		memcpy(ptr, &hdr, sizeof(hdr));
+		ptr += sizeof(hdr);
+		memcpy(ptr, data, bytes);
 
 		queued_data.push_back(data_element);
 	} else if (encoding == Stream::STREAM_ENCODING_RAW) {
@@ -275,7 +325,7 @@ void Stream::process_queued_data()
 	// Hold the lock for as short as possible, since add_data_raw() can possibly
 	// write to disk, which might disturb the input thread.
 	{
-		MutexLock lock(&queued_data_mutex);
+		lock_guard<mutex> lock(queued_data_mutex);
 		if (queued_data.empty()) {
 			return;
 		}
@@ -292,8 +342,9 @@ void Stream::process_queued_data()
 	// data, and 10 kB is a very fine granularity in most streams.
 	static const int minimum_start_point_distance = 10240;
 	size_t byte_position = bytes_received;
-	for (size_t i = 0; i < queued_data_copy.size(); ++i) {
-		if ((queued_data_copy[i].metacube_flags & METACUBE_FLAGS_NOT_SUITABLE_FOR_STREAM_START) == 0) {
+	bool need_hls_clear = false;
+	for (const DataElement &elem : queued_data_copy) {
+		if ((elem.metacube_flags & METACUBE_FLAGS_NOT_SUITABLE_FOR_STREAM_START) == 0) {
 			size_t num_points = suitable_starting_points.size();
 			if (num_points >= 2 &&
 			    suitable_starting_points[num_points - 1] - suitable_starting_points[num_points - 2] < minimum_start_point_distance) {
@@ -301,14 +352,21 @@ void Stream::process_queued_data()
 				suitable_starting_points.pop_back();
 			}
 			suitable_starting_points.push_back(byte_position);
+
+			if (elem.pts.timebase_num != 0) {
+				need_hls_clear |= add_fragment_boundary(byte_position, elem.pts);
+			}
 		}
-		byte_position += queued_data_copy[i].data.iov_len;
+		byte_position += elem.data.iov_len;
+	}
+	if (need_hls_clear) {
+		clear_hls_playlist_cache();
 	}
 
 	add_data_raw(queued_data_copy);
 	remove_obsolete_starting_points();
-	for (size_t i = 0; i < queued_data_copy.size(); ++i) {
-		char *data = reinterpret_cast<char *>(queued_data_copy[i].data.iov_base);
+	for (const DataElement &elem : queued_data_copy) {
+		char *data = reinterpret_cast<char *>(elem.data.iov_base);
 		delete[] data;
 	}
 
@@ -320,3 +378,96 @@ void Stream::process_queued_data()
 		sleeping_clients.clear();
 	}
 }
+
+bool Stream::add_fragment_boundary(size_t byte_position, const RationalPTS &pts)
+{
+	double pts_double = double(pts.pts) * pts.timebase_den / pts.timebase_num;
+
+	if (fragments.size() <= 1) {
+		// Just starting up, so try to establish the first in-progress fragment.
+		fragments.push_back(FragmentStart{ byte_position, pts_double });
+		return false;
+	}
+
+	// Keep extending the in-progress fragment as long as we do not
+	// exceed the target duration by more than half a second
+	// (RFC 8216 4.3.3.1) and we get closer to the target by doing so.
+	// Note that in particular, this means we'll always extend
+	// as long as we don't exceed the target duration.
+	double current_duration = fragments[fragments.size() - 1].pts;
+	double candidate_duration = pts_double - fragments[fragments.size() - 2].pts;
+	if (lrintf(candidate_duration) <= hls_frag_duration &&
+	    fabs(candidate_duration - hls_frag_duration) < fabs(current_duration - hls_frag_duration)) {
+		fragments.back() = FragmentStart{ byte_position, pts_double };
+		return false;
+	} else {
+		// Extending the in-progress fragment would make it too long,
+		// so finalize it and start a new in-progress fragment.
+		fragments.push_back(FragmentStart{ byte_position, pts_double });
+		return true;
+	}
+}
+
+void Stream::clear_hls_playlist_cache()
+{
+	hls_playlist_http10.reset();
+	hls_playlist_http11_close.reset();
+	hls_playlist_http11_persistent.reset();
+}
+
+shared_ptr<const string> Stream::generate_hls_playlist(bool http_11, bool close_after_response)
+{
+	char buf[256];
+	snprintf(buf, sizeof(buf),
+		"#EXTM3U\r\n"
+		"#EXT-X-VERSION:7\r\n"
+		"#EXT-X-TARGETDURATION:%u\r\n"
+		"#EXT-X-MEDIA-SEQUENCE:%" PRIu64 "\r\n"
+		"#EXT-X-DISCONTINUITY-SEQUENCE:%" PRIu64 "\r\n",
+		hls_frag_duration,
+		first_fragment_index,
+		discontinuity_counter);
+
+	string playlist = buf;
+
+	if (!stream_header.empty()) {
+		snprintf(buf, sizeof(buf), "#EXT-X-MAP:URI=\"%s?frag=header\"\r\n", url.c_str());
+		playlist += buf;
+	}
+
+	playlist += "\r\n";
+	if (fragments.size() >= 3) {
+		for (size_t i = 0; i < fragments.size() - 2; ++i) {
+			char buf[256];
+			snprintf(buf, sizeof(buf), "#EXTINF:%f,\r\n%s?frag=%" PRIu64 "-%" PRIu64 "\r\n",
+				fragments[i + 1].pts - fragments[i].pts,
+				url.c_str(),
+				fragments[i].byte_position,
+				fragments[i + 1].byte_position);
+			playlist += buf;
+		}
+	}
+
+	string response;
+	if (http_11) {
+		response = "HTTP/1.1 200 OK\r\n";
+		if (close_after_response) {
+			response.append("Connection: close\r\n");
+		}
+	} else {
+		assert(close_after_response);
+		response = "HTTP/1.0 200 OK\r\n";
+	}
+	snprintf(buf, sizeof(buf), "Content-Length: %zu\r\n", playlist.size());
+	response.append(buf);
+	response.append("Content-Type: application/x-mpegURL\r\n");
+	if (!allow_origin.empty()) {
+		response.append("Access-Control-Allow-Origin: ");
+		response.append(allow_origin);
+		response.append("\r\n");
+	}
+	response.append("\r\n");
+	response.append(move(playlist));
+
+	return shared_ptr<const string>(new string(move(response)));
+}
